import pathlib

import pydantic
from truss.templates.shared import secrets_resolver
from truss_chains import definitions

# Better: in >=3.10 use `TypeAlias`.
UserConfigT = pydantic.BaseModel


class TrussChainletModel:
    _context: definitions.DeploymentContext[UserConfigT]
    _chainlet: definitions.ABCChainlet

    def __init__(
        self, config: dict, data_dir: pathlib.Path, secrets: secrets_resolver.Secrets
    ) -> None:
        truss_metadata: definitions.TrussMetadata[
            UserConfigT
        ] = definitions.TrussMetadata[UserConfigT].parse_obj(
            config["model_metadata"][definitions.TRUSS_CONFIG_CHAINS_KEY]
        )
        self._context = definitions.DeploymentContext[UserConfigT](
            user_config=truss_metadata.user_config,
            chainlet_to_service=truss_metadata.chainlet_to_service,
            secrets=secrets,
        )

    # Below illustrated code will be added by code generation.

    # def load(self):
    #     self._chainlet = {ChainletCls}(self._context)

    # Sync or async.
    # def predict(self, payload):
    #     return self._chainlet.{method_name}(payload)



# flake8: noqa F401
from truss_chains.definitions import (
    Assets,
    Compute,
    DeploymentContext,
    DockerImage,
    RemoteConfig,
    UsageError,
)
from truss_chains.public_api import (
    ChainletBase,
    deploy_remotely,
    provide,
    provide_context,
    run_local,
)
from truss_chains.utils import make_abs_path_here



import ast
import json
import logging
import pathlib
import shlex
import subprocess
import textwrap
from typing import Any, Iterable, Mapping, Optional, cast

import datamodel_code_generator
import libcst
from datamodel_code_generator import model as generator_models
from datamodel_code_generator.parser import jsonschema as jsonschema_parser
from truss_chains import definitions, model_skeleton, utils

INDENT = " " * 4

_DEFS_KEY = "$defs"


def _indent(text: str, num: int = 1) -> str:
    return textwrap.indent(text, INDENT * num)


def _run_simple_subprocess(cmd: str) -> None:
    process = subprocess.Popen(
        shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    _, stderr = process.communicate()
    if process.returncode != 0:
        raise ChildProcessError(f"Error: {stderr.decode()}")


def _format_python_file(file_path: pathlib.Path) -> None:
    _run_simple_subprocess(
        f"autoflake --in-place --remove-all-unused-imports {file_path}"
    )
    _run_simple_subprocess(f"black {file_path}")
    _run_simple_subprocess(f"isort {file_path}")


def make_chainlet_dir(
    chain_name: str,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    root: pathlib.Path = pathlib.Path("/tmp"),
) -> pathlib.Path:
    chainlet_name = chainlet_descriptor.name
    file_name = f"chainlet_{chainlet_name}"
    chainlet_dir = root / definitions.GENERATED_CODE_DIR / chain_name / file_name
    chainlet_dir.mkdir(exist_ok=True, parents=True)
    return chainlet_dir


# Truss Gen ############################################################################


class _SpecifyChainletTypeAnnotation(libcst.CSTTransformer):
    def __init__(self, new_annotation: str) -> None:
        super().__init__()
        self._new_annotation = new_annotation

    def leave_SimpleStatementLine(
        self,
        original_node: libcst.SimpleStatementLine,
        updated_node: libcst.SimpleStatementLine,
    ) -> libcst.SimpleStatementLine:
        new_body: list[Any] = []
        for statement in updated_node.body:
            if (
                isinstance(statement, libcst.AnnAssign)
                and isinstance(statement.target, libcst.Name)
                and statement.target.value == "_chainlet"
            ):
                new_annotation = libcst.Annotation(
                    annotation=libcst.Name(value=self._new_annotation)
                )
                new_statement = statement.with_changes(annotation=new_annotation)
                new_body.append(new_statement)
            else:
                new_body.append(statement)

        return updated_node.with_changes(body=tuple(new_body))


def _gen_load_src(chainlet_descriptor: definitions.ChainletAPIDescriptor):
    """Generates AST for the `load` method of the truss model."""
    stub_inits = []
    stub_args = []
    for name, proc_cls in chainlet_descriptor.dependencies.items():
        cls_name = utils.make_stub_name(proc_cls.__name__)
        stub_inits.append(
            f"{name} = stub.factory({cls_name}, self._context, '{proc_cls.__name__}')"
        )
        stub_args.append(f"{name}={name}")

    if stub_args:
        init_args = f"context=self._context, {', '.join(stub_args)}"
    else:
        init_args = "context=self._context"

    body = _indent(
        "\n".join(
            [f"logging.info(f'Loading Chainlet `{chainlet_descriptor.name}`.')"]
            + stub_inits
            + [f"self._chainlet = {chainlet_descriptor.name}({init_args})"]
        )
    )
    return libcst.parse_statement("\n".join(["def load(self) -> None:", body]))


def _gen_predict_src(
    endpoint_descriptor: definitions.EndpointAPIDescriptor, chainlet_name: str
):
    """Generates AST for the `predict` method of the truss model."""
    if endpoint_descriptor.is_generator:
        # TODO: implement generator.
        raise NotImplementedError("Generator.")

    parts = []
    def_str = "async def" if endpoint_descriptor.is_async else "def"
    parts.append(f"{def_str} predict(self, payload):")
    # Add error handling context manager:
    parts.append(
        _indent(
            f"with utils.exception_to_http_error("
            f'include_stack=True, chainlet_name="{chainlet_name}"):'
        )
    )
    # Convert items from json payload dict to an arg-list, parsing pydantic models.
    args = ", ".join(
        (
            f"{arg_name}={arg_type.as_src_str()}.parse_obj(payload['{arg_name}'])"
            if arg_type.is_pydantic
            else f"{arg_name}=payload['{arg_name}']"
        )
        for arg_name, arg_type in endpoint_descriptor.input_names_and_types
    )
    # Invoke Chainlet.
    maybe_await = "await " if endpoint_descriptor.is_async else ""
    parts.append(
        _indent(
            f"result = {maybe_await}self._chainlet.{endpoint_descriptor.name}({args})",
            2,
        )
    )
    # Return as json tuple, serialize pydantic models.
    if len(endpoint_descriptor.output_types) == 1:
        output_type = endpoint_descriptor.output_types[0]
        result = "result.dict()" if output_type.is_pydantic else "result"
    else:
        result_parts = [
            f"result[{i}].dict()" if t.is_pydantic else f"result[{i}]"
            for i, t in enumerate(endpoint_descriptor.output_types)
        ]
        result = f"{', '.join(result_parts)}"

    parts.append(_indent(f"return {result}"))

    return libcst.parse_statement("\n".join(parts))


def _generate_truss_model(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
) -> tuple[libcst.CSTNode, list[libcst.SimpleStatementLine], libcst.CSTNode]:
    logging.info(f"Generating Truss model for `{chainlet_descriptor.name}`.")
    skeleton_tree = libcst.parse_module(
        pathlib.Path(model_skeleton.__file__).read_text()
    )
    imports: list[Any] = [
        node
        for node in skeleton_tree.body
        if isinstance(node, libcst.SimpleStatementLine)
        and any(
            isinstance(stmt, libcst.Import) or isinstance(stmt, libcst.ImportFrom)
            for stmt in node.body
        )
    ]
    imports.append(libcst.parse_statement("import logging"))
    imports.append(libcst.parse_statement("from truss_chains import utils"))

    class_definition: libcst.ClassDef = utils.expect_one(
        node
        for node in skeleton_tree.body
        if isinstance(node, libcst.ClassDef)
        and node.name.value == model_skeleton.TrussChainletModel.__name__
    )

    load_def = _gen_load_src(chainlet_descriptor)
    predict_def = _gen_predict_src(
        chainlet_descriptor.endpoint, chainlet_descriptor.name
    )
    new_body: list[Any] = list(class_definition.body.body) + [load_def, predict_def]
    new_block = libcst.IndentedBlock(body=new_body)
    class_definition = class_definition.with_changes(body=new_block)
    class_definition = class_definition.visit(  # type: ignore[assignment]
        _SpecifyChainletTypeAnnotation(chainlet_descriptor.name)
    )
    if issubclass(chainlet_descriptor.user_config_type.raw, type(None)):
        userconfig_pin = libcst.parse_statement("UserConfigT = type(None)")
    else:
        userconfig_pin = libcst.parse_statement(
            f"UserConfigT = {chainlet_descriptor.user_config_type.as_src_str()}"
        )
    return class_definition, imports, userconfig_pin


# Stub Gen #############################################################################


def _endpoint_signature_src(endpoint: definitions.EndpointAPIDescriptor):
    """
    E.g.: `async def run(self, data: str, num_partitions: int) -> tuple[list, int]:`
    """
    if endpoint.is_generator:
        # TODO: implement generator.
        raise NotImplementedError("Generator.")

    def_str = "async def" if endpoint.is_async else "def"
    args = ", ".join(
        f"{arg_name}: {arg_type.as_src_str(definitions.STUB_TYPE_MODULE)}"
        for arg_name, arg_type in endpoint.input_names_and_types
    )
    if len(endpoint.output_types) == 1:
        output_type = (
            f"{endpoint.output_types[0].as_src_str(definitions.STUB_TYPE_MODULE)}"
        )
    else:
        out_types = ", ".join(
            t.as_src_str(definitions.STUB_TYPE_MODULE) for t in endpoint.output_types
        )
        output_type = f"tuple[{out_types}]"
    return f"{def_str} {endpoint.name}(self, {args}) -> {output_type}:"


def _endpoint_body_src(endpoint: definitions.EndpointAPIDescriptor) -> str:
    """Generates source code for calling the stub and wrapping the I/O types.

    E.g.:
    ```
    json_args = {"inputs": inputs.dict(), "extra_arg": extra_arg}
    json_result = await self._remote.predict_async(json_args)
    return (SplitTextOutput.parse_obj(json_result[0]), json_result[1])
    ```
    """
    if endpoint.is_generator:
        raise NotImplementedError("Generator")

    parts = []
    # Pack arg list as json dictionary.
    json_arg_parts = (
        (
            f"'{arg_name}': {arg_name}.dict()"
            if arg_type.is_pydantic
            else f"'{arg_name}': {arg_name}"
        )
        for arg_name, arg_type in endpoint.input_names_and_types
    )
    json_args = f"{{{', '.join(json_arg_parts)}}}"
    parts.append(f"json_args = {json_args}")
    # Invoke remote.
    remote_call = (
        "await self._remote.predict_async(json_args)"
        if endpoint.is_async
        else "self._remote.predict_sync(json_args)"
    )
    parts.append(f"json_result = {remote_call}")
    # Unpack response and parse as pydantic models if needed.
    if len(endpoint.output_types) == 1:
        output_type = utils.expect_one(endpoint.output_types)
        if output_type.is_pydantic:
            type_str = output_type.as_src_str(definitions.STUB_TYPE_MODULE)
            ret = f"{type_str}.parse_obj(json_result)"
        else:
            ret = "json_result"
    else:
        ret_parts = ", ".join(
            (
                (
                    f"{output_type.as_src_str(definitions.STUB_TYPE_MODULE)}"
                    f".parse_obj(json_result[{i}])"
                )
                if output_type.is_pydantic
                else f"json_result[{i}]"
            )
            for i, output_type in enumerate(endpoint.output_types)
        )
        ret = f"{ret_parts}"
    parts.append(f"return {ret}")

    return "\n".join(parts)


def _gen_stub_src(
    Chainlet: definitions.ChainletAPIDescriptor,
    maybe_pydantic_types_file: Optional[pathlib.Path],
) -> tuple[str, list[str]]:
    """Generates stub class source, e.g:

    ```
    from truss_chains import stub

    class SplitText(stub.StubBase):
        def __init__(self, url: str, api_key: str) -> None:
            self._remote = stub.BasetenSession(url, api_key)

        async def run(self, data: str, num_partitions: int) -> tuple[SplitTextOutput, int]:
            json_args = {"inputs": inputs.dict(), "extra_arg": extra_arg}
            json_result = await self._remote.predict_async(json_args)
            return (SplitTextOutput.parse_obj(json_result[0]), json_result[1])
    ```
    """
    imports = ["from truss_chains import stub"]
    if maybe_pydantic_types_file:
        imports.append(f"import {definitions.STUB_TYPE_MODULE}")

    src_parts = [
        f"class {utils.make_stub_name(Chainlet.name)}(stub.StubBase):",
        _indent(_endpoint_signature_src(Chainlet.endpoint)),
        _indent(_endpoint_body_src(Chainlet.endpoint), 2),
        "\n\n",
    ]
    return "\n".join(src_parts), imports


def _remove_root_model(pydantic_src: str) -> str:
    """To process source code generated by `datamodel_code_generator`."""
    tree = ast.parse(pydantic_src)
    src = []
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == "Model":
            continue
        else:
            src.append(node)

    return f"{ast.unparse(ast.Module(body=src, type_ignores=[]))}\n\n"


def _export_pydantic_schemas(
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
) -> Mapping[str, Any]:
    """Creates a dict with all pydantic schemas used as input or output types by the
    dependencies.

    If a schema itself depends on another class (e.g. an enum definition), these
    dependent schemas are also added to the dict.

    It is enforced that pydantic models across the code base have different names,
    to avoid conflicts; automatic disambiguation is not yet implemented.
    """
    name_to_schema: dict[str, Any] = {}

    def safe_add_schema(type_descr: definitions.TypeDescriptor):
        if not type_descr.is_pydantic:
            return

        name = type_descr.raw.__name__
        schema = type_descr.raw.schema()
        if existing_schema := name_to_schema.get(name):
            if existing_schema != schema:
                raise NotImplementedError(
                    f"Two pydantic models with same name `{name}."
                )
        else:
            # Move definitions needed by the current type to top level.
            if _DEFS_KEY in schema:
                for def_name, def_schema in schema[_DEFS_KEY].items():
                    if existing_def_schema := name_to_schema.get(def_name):
                        if existing_def_schema != def_schema:
                            raise NotImplementedError(
                                f"Two pydantic models with same name `{def_name}."
                            )
                    else:
                        name_to_schema[def_name] = def_schema
                schema.pop(_DEFS_KEY)

            name_to_schema[name] = schema

    for dep in dependencies:
        for _, input_type in dep.endpoint.input_names_and_types:
            safe_add_schema(input_type)
        for output_type in dep.endpoint.output_types:
            safe_add_schema(output_type)
    return name_to_schema


def gen_pydantic_models(
    file_path: pathlib.Path,
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
) -> Optional[pathlib.Path]:
    name_to_schema = _export_pydantic_schemas(dependencies)

    if not name_to_schema:
        return None

    combined_schema = {_DEFS_KEY: name_to_schema}
    # TODO: parameterize pydantic and python version properly.
    py_version = datamodel_code_generator.PythonVersion.PY_39
    data_model_types = generator_models.get_data_model_types(
        data_model_type=datamodel_code_generator.DataModelType.PydanticBaseModel,
        target_python_version=py_version,
    )
    with utils.log_level(logging.INFO):
        parser = jsonschema_parser.JsonSchemaParser(
            json.dumps(combined_schema),
            data_model_type=data_model_types.data_model,
            data_model_root_type=data_model_types.root_model,
            data_model_field_type=data_model_types.field_model,
            data_type_manager_type=data_model_types.data_type_manager,
            target_python_version=py_version,
            capitalise_enum_members=True,
            use_subclass_enum=True,
            strict_nullable=True,
            apply_default_values_for_required_fields=True,
            use_default_kwarg=True,
            use_standard_collections=True,
        )
        pydantic_module_src: str = cast(str, parser.parse())
    file_path.write_text(_remove_root_model(pydantic_module_src))
    return file_path


def _generate_stub_src_for_deps(
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
    pydantic_types_file: Optional[pathlib.Path],
) -> Optional[tuple[str, set[str]]]:
    """Generates a source code and imports for stub classes."""
    imports = set()
    src_parts = []

    for dep in dependencies:
        stub_src, new_imports = _gen_stub_src(dep, pydantic_types_file)
        imports.update(new_imports)
        src_parts.append(stub_src)

    if not (imports or src_parts):
        return None
    return "\n".join(src_parts), imports


# Remote Chainlet Gen #################################################################


class _MainRemover(ast.NodeTransformer):
    """Removes main-section from module AST."""

    def visit_If(self, node):
        """Robustly matches variations of `if __name__ == "__main__":`."""
        if (
            isinstance(node.test, ast.Compare)
            and any(
                isinstance(c, ast.Name) and c.id == "__name__"
                for c in ast.walk(node.test.left)
            )
            and any(
                isinstance(c, ast.Constant) and c.value == "__main__"
                for c in ast.walk(node.test)
            )
        ):
            return None
        return self.generic_visit(node)


def _remove_main_section(source_code: str) -> str:
    """Removes main-section from module source."""
    parsed_code = ast.parse(source_code)
    transformer = _MainRemover()
    transformed_ast = transformer.visit(parsed_code)
    return ast.unparse(transformed_ast)


def generate_chainlet_source(
    file_path: pathlib.Path,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
    pydantic_types_file: Optional[pathlib.Path],
):
    """Generates code that wraps a Chainlet as a truss-compatible model."""
    maybe_stub_src = _generate_stub_src_for_deps(dependencies, pydantic_types_file)

    source_code = _remove_main_section(file_path.read_text())
    if maybe_stub_src:
        stub_src, stub_imports = maybe_stub_src
        stub_imports_str = "\n".join(stub_imports)
        source_code = f"{stub_imports_str}\n{source_code}\n{stub_src}"

    source_tree = libcst.parse_module(source_code)

    # TODO: Chainlet isolation: either prune file from high-confidence unneeded.

    model_def, imports, userconfig_pin = _generate_truss_model(chainlet_descriptor)
    new_body: list[Any] = imports + list(source_tree.body) + [userconfig_pin, model_def]
    source_tree = source_tree.with_changes(body=new_body)
    file_path.write_text(source_tree.code)
    _format_python_file(file_path)



from typing import Optional

import numpy as np
import tritonclient
import tritonclient.grpc.aio as grpcclient


class ModelInput:
    def __init__(
        self,
        prompt: str,
        request_id: int,
        max_tokens: int = 50,
        max_new_tokens: int = 50,
        temperature: float = 1.0,
        top_p: float = 1.0,
        top_k: int = 50,
        beam_width: int = 1,
        bad_words_list: Optional[list] = None,
        stop_words_list: Optional[list] = None,
        repetition_penalty: float = 1.0,
        ignore_eos: bool = False,
        stream: bool = True,
        eos_token_id: int = None,  # type: ignore
    ) -> None:
        self.stream = stream
        self.request_id = request_id
        self._prompt = prompt
        self._max_tokens = max_tokens
        self._beam_width = beam_width
        self._bad_words_list = [""] if bad_words_list is None else bad_words_list
        self._stop_words_list = [""] if stop_words_list is None else stop_words_list
        self._repetition_penalty = repetition_penalty
        self._eos_token_id = eos_token_id
        self._ignore_eos = ignore_eos
        # These variables are passed by OAI proxy but are unused
        # TODO(Abu): Add support for these
        self._max_new_tokens = max_new_tokens
        self._temperature = temperature
        self._top_p = top_p
        self._top_k = top_k

    def _prepare_grpc_tensor(
        self, name: str, input_data: np.ndarray
    ) -> grpcclient.InferInput:
        tensor = grpcclient.InferInput(
            name,
            input_data.shape,
            tritonclient.utils.np_to_triton_dtype(input_data.dtype),
        )
        tensor.set_data_from_numpy(input_data)
        return tensor

    def to_tensors(self):
        if self._eos_token_id is None and self._ignore_eos:
            raise ValueError("eos_token_id is required when ignore_eos is True")

        prompt_data = np.array([[self._prompt]], dtype=object)
        output_len_data = np.ones_like(prompt_data, dtype=np.uint32) * self._max_tokens
        bad_words_data = np.array([self._bad_words_list], dtype=object)
        stop_words_data = np.array([self._stop_words_list], dtype=object)
        stream_data = np.array([[self.stream]], dtype=bool)
        beam_width_data = np.array([[self._beam_width]], dtype=np.uint32)
        repetition_penalty_data = np.array(
            [[self._repetition_penalty]], dtype=np.float32
        )

        inputs = [
            self._prepare_grpc_tensor("text_input", prompt_data),
            self._prepare_grpc_tensor("max_tokens", output_len_data),
            self._prepare_grpc_tensor("bad_words", bad_words_data),
            self._prepare_grpc_tensor("stop_words", stop_words_data),
            self._prepare_grpc_tensor("stream", stream_data),
            self._prepare_grpc_tensor("beam_width", beam_width_data),
            self._prepare_grpc_tensor("repetition_penalty", repetition_penalty_data),
        ]

        if not self._ignore_eos:
            end_id_data = np.array([[self._eos_token_id]], dtype=np.uint32)
            inputs.append(self._prepare_grpc_tensor("end_id", end_id_data))

        return inputs


# The following are duplicated from the underlying base image.
# We list them as a comment for posterity:
#
# class TRTLLMModelArchitecture(Enum):
#     LLAMA: str = "llama"
#     MISTRAL: str = "mistral"
#     DEEPSEEK: str = "deepseek"


# class TRTLLMQuantizationType(Enum):
#     NO_QUANT: str = "no_quant"
#     WEIGHTS_ONLY_INT8: str = "weights_int8"
#     WEIGHTS_KV_INT8: str = "weights_kv_int8"
#     WEIGHTS_ONLY_INT4: str = "weights_int4"
#     WEIGHTS_KV_INT4: str = "weights_kv_int4"
#     SMOOTH_QUANT: str = "smooth_quant"
#     FP8: str = "fp8"
#     FP8_KV: str = "fp8_kv"

# class TrussTRTLLMPluginConfiguration(BaseModel):
#     multi_block_mode: bool = False
#     paged_kv_cache: bool = True
#     use_fused_mlp: bool = False

# class TrussTRTLLMBuildConfiguration(BaseModel):
#     base_model_architecture: TRTLLMModelArchitecture
#     max_input_len: int
#     max_output_len: int
#     max_batch_size: int
#     max_beam_width: int
#     max_prompt_embedding_table_size: int = 0
#     huggingface_ckpt_repository: Optional[str]
#     gather_all_token_logits: bool = False
#     strongly_typed: bool = False
#     quantization_type: TRTLLMQuantizationType = TRTLLMQuantizationType.NO_QUANT
#     tensor_parallel_count: int = 1
#     pipeline_parallel_count: int = 1
#     plugin_configuration: TrussTRTLLMPluginConfiguration = TrussTRTLLMPluginConfiguration()

# class TrussTRTLLMServingConfiguration(BaseModel):
#     engine_repository: str
#     tokenizer_repository: str
#     tensor_parallel_count: int = 1
#     pipeline_parallel_count: int = 1

# class TrussTRTLLMConfiguration(BaseModel):
#     serve: Optional[TrussTRTLLMServingConfiguration] = None
#     build: Optional[TrussTRTLLMBuildConfiguration] = None

#     @model_validator(mode="after")
#     def check_minimum_required_configuration(self):
#         if not self.serve and not self.build:
#             raise ValueError(
#                 "Either serve or build configurations must be provided"
#             )
#         if self.serve and self.build:
#             raise ValueError(
#                 "Both serve and build configurations cannot be provided"
#             )
#         if self.serve is not None:
#             if (self.serve.engine_repository is None) ^ (self.serve.tokenizer_repository is None):
#                 raise ValueError(
#                     "Both engine_repository and tokenizer_repository must be provided"
#                 )
#         return self

#     @property
#     def requires_build(self):
#         if self.build is not None:
#             return True
#         return False



# TODO: this file contains too much implementation -> restructure.

import abc
import logging
import os
import traceback
from types import GenericAlias
from typing import Any, ClassVar, Generic, Mapping, Optional, Type, TypeVar, Union

import pydantic
from pydantic import generics
from truss import truss_config

UserConfigT = TypeVar("UserConfigT", bound=Optional[pydantic.BaseModel])

BASETEN_API_SECRET_NAME = "baseten_chain_api_key"
SECRET_DUMMY = "***"
TRUSS_CONFIG_CHAINS_KEY = "chains_metadata"

ENDPOINT_METHOD_NAME = "run"  # Referring to Chainlet method name exposed as endpoint.
# Below arg names must correspond to `definitions.ABCChainlet`.
CONTEXT_ARG_NAME = "context"  # Referring to Chainlets `__init__` signature.
SELF_ARG_NAME = "self"

GENERATED_CODE_DIR = ".chains_generated"
PREDICT_ENDPOINT_NAME = "/predict"
chainlet_MODULE = "Chainlet"
STUB_TYPE_MODULE = "stub_types"
STUB_CLS_SUFFIX = "Stub"


class APIDefinitionError(TypeError):
    """Raised when user-defined Chainlets do not adhere to API constraints."""


class MissingDependencyError(TypeError):
    """Raised when a needed resource could not be found or is not defined."""


class UsageError(Exception):
    """Raised when components are not used the expected way at runtime."""


class AbsPath:
    _abs_file_path: str
    _creating_module: str
    _original_path: str

    def __init__(
        self, abs_file_path: str, creating_module: str, original_path: str
    ) -> None:
        self._abs_file_path = abs_file_path
        self._creating_module = creating_module
        self._original_path = original_path

    def raise_if_not_exists(self) -> None:
        if not os.path.isfile(self._abs_file_path):
            raise MissingDependencyError(
                f"With the file path `{self._original_path}` an absolute path relative "
                f"to the calling module `{self._creating_module}` was created, "
                f"resulting `{self._abs_file_path}` - but no file was found."
            )

    @property
    def abs_path(self) -> str:
        if self._abs_file_path != self._original_path:
            logging.info(
                f"Using abs path `{self._abs_file_path}` for path specified as "
                f"`{self._original_path}` (in `{self._creating_module}`)."
            )
        return self._abs_file_path


class DockerImageSpec(pydantic.BaseModel):
    class Config:
        arbitrary_types_allowed = True

    # TODO: this is not stable yet and might change or refer back to truss.
    base_image: str = "python:3.11-slim"
    pip_requirements_file: Optional[AbsPath] = None
    pip_requirements: list[str] = []
    apt_requirements: list[str] = []


class DockerImage:
    """Builder to create ImageSpec."""

    _spec: DockerImageSpec

    def __init__(self) -> None:
        self._spec = DockerImageSpec()

    def base_image(self, base_image: str) -> "DockerImage":
        self._spec.base_image = base_image
        return self

    def pip_requirements_file(self, file_path: AbsPath) -> "DockerImage":
        self._spec.pip_requirements_file = file_path
        return self

    def pip_requirements(self, requirements: list[str]) -> "DockerImage":
        self._spec.pip_requirements = requirements
        return self

    def apt_requirements(self, requirements: list[str]) -> "DockerImage":
        self._spec.apt_requirements = requirements
        return self

    def get_spec(self) -> DockerImageSpec:
        return self._spec.copy(deep=True)


class ComputeSpec(pydantic.BaseModel):
    # TODO: this is not stable yet and might change or refer back to truss.
    cpu: int = 1
    memory: str = "2Gi"
    gpu: truss_config.AcceleratorSpec = truss_config.AcceleratorSpec()


class Compute:
    """Builder to create ComputeSpec."""

    _spec: ComputeSpec

    def __init__(self) -> None:
        self._spec = ComputeSpec()

    def cpu(self, cpu: int) -> "Compute":
        self._spec.cpu = cpu
        return self

    def memory(self, memory: str) -> "Compute":
        self._spec.memory = memory
        return self

    def gpu(
        self, kind: Union[str, truss_config.Accelerator], count: int = 1
    ) -> "Compute":
        self._spec.gpu = truss_config.AcceleratorSpec(
            accelerator=truss_config.Accelerator(kind), count=count
        )
        return self

    def get_spec(self) -> ComputeSpec:
        return self._spec.copy(deep=True)


class AssetSpec(pydantic.BaseModel):
    # TODO: this is not stable yet and might change or refer back to truss.
    secrets: dict[str, str] = {}
    cached: list[Any] = []


class Assets:
    """Builder to create asset spec."""

    _spec: AssetSpec

    def __init__(self) -> None:
        self._spec = AssetSpec()

    def add_secret(self, key: str) -> "Assets":
        # Actual value is provided in deployment.
        self._spec.secrets[key] = SECRET_DUMMY
        return self

    def cached(self, value: list[Any]) -> "Assets":
        self._spec.cached = value
        return self

    def get_spec(self) -> AssetSpec:
        return self._spec.copy(deep=True)


class RemoteConfig(pydantic.BaseModel):
    """Bundles config values needed to deploy a Chainlet."""

    class Config:
        arbitrary_types_allowed = True

    docker_image: DockerImage
    compute: Compute = Compute()
    assets: Assets = Assets()
    name: Optional[str] = None

    def get_docker_image_spec(self) -> DockerImageSpec:
        return self.docker_image.get_spec()

    def get_compute_spec(self) -> ComputeSpec:
        return self.compute.get_spec()

    def get_asset_spec(self) -> AssetSpec:
        return self.assets.get_spec()


class ServiceDescriptor(pydantic.BaseModel):
    name: str
    predict_url: str


class DeploymentContext(generics.GenericModel, Generic[UserConfigT]):
    """Bundles config values and resources needed to instantiate Chainlets."""

    class Config:
        arbitrary_types_allowed = True

    user_config: UserConfigT = pydantic.Field(default=None)
    chainlet_to_service: Mapping[str, ServiceDescriptor] = {}
    # secrets: Optional[secrets_resolver.Secrets] = None
    # TODO: above type results in `truss.server.shared.secrets_resolver.Secrets`
    #   due to the templating, at runtime the object passed will be from
    #   `shared.secrets_resolver` and give pydantic validation error.
    secrets: Optional[Any] = None

    def get_service_descriptor(self, chainlet_name: str) -> ServiceDescriptor:
        if chainlet_name not in self.chainlet_to_service:
            raise MissingDependencyError(f"{chainlet_name}")
        return self.chainlet_to_service[chainlet_name]

    def get_baseten_api_key(self) -> str:
        if self.secrets is None:
            raise UsageError(f"Secrets not set in `{self.__class__.__name__}` object.")
        error_msg = (
            "For using chains, it is required to setup a an API key with name "
            f"`{BASETEN_API_SECRET_NAME}` on baseten to allow chain Chainlet to "
            "call other Chainlets. For local execution, secrets can be provided "
            "to `run_local`."
        )
        if BASETEN_API_SECRET_NAME not in self.secrets:
            raise MissingDependencyError(error_msg)

        api_key = self.secrets[BASETEN_API_SECRET_NAME]
        if api_key == SECRET_DUMMY:
            raise MissingDependencyError(
                f"{error_msg}. Retrieved dummy value of `{api_key}`."
            )
        return api_key


class TrussMetadata(generics.GenericModel, Generic[UserConfigT]):
    """Plugin for the truss config (in config["model_metadata"]["chains_metadata"])."""

    user_config: UserConfigT = pydantic.Field(default=None)
    chainlet_to_service: Mapping[str, ServiceDescriptor] = {}


class ABCChainlet(Generic[UserConfigT], abc.ABC):
    remote_config: ClassVar[RemoteConfig]
    default_user_config: ClassVar[Optional[pydantic.BaseModel]] = None
    _init_is_patched: ClassVar[bool] = False

    _context: DeploymentContext[UserConfigT]

    @abc.abstractmethod
    def __init__(self, context: DeploymentContext[UserConfigT]) -> None:
        ...

    # Cannot add this abstract method to API, because we want to allow arbitrary
    # arg/kwarg names and specifying any function signature here would give type errors
    # @abc.abstractmethod
    # def run(self, *args, **kwargs) -> Any: ...

    @property
    @abc.abstractmethod
    def user_config(self) -> UserConfigT:
        ...


class TypeDescriptor(pydantic.BaseModel):
    """For describing I/O types of Chainlets.

    Example:
        repr(raw): <class 'user_package.shared_chainlet.SplitTextInput'>"
        as_src_str(): 'SplitTextInput'

    -> Source string, without further qualification, does not include any module path.
    """

    raw: Any  # The raw type annotation object (could be a type or GenericAlias).

    def as_src_str(self, qualify_pydantic_types: Optional[str] = None) -> str:
        # TODO: pydantic types will soon be handled differently.
        if self.is_pydantic and qualify_pydantic_types:
            return f"{qualify_pydantic_types}.{self.raw.__name__}"

        if isinstance(self.raw, type):
            return self.raw.__name__
        else:
            return str(self.raw)

    @property
    def is_pydantic(self) -> bool:
        return (
            isinstance(self.raw, type)
            and not isinstance(self.raw, GenericAlias)
            and issubclass(self.raw, pydantic.BaseModel)
        )


class EndpointAPIDescriptor(pydantic.BaseModel):
    name: str = ENDPOINT_METHOD_NAME
    input_names_and_types: list[tuple[str, TypeDescriptor]]
    output_types: list[TypeDescriptor]
    is_async: bool
    is_generator: bool


class ChainletAPIDescriptor(pydantic.BaseModel):
    chainlet_cls: Type[ABCChainlet]
    src_path: str
    dependencies: Mapping[str, Type[ABCChainlet]]
    endpoint: EndpointAPIDescriptor
    user_config_type: TypeDescriptor

    def __hash__(self) -> int:
        return hash(self.chainlet_cls)

    @property
    def name(self) -> str:
        return self.chainlet_cls.__name__


class StackFrame(pydantic.BaseModel):
    filename: str
    lineno: Optional[int]
    name: str
    line: Optional[str]

    @classmethod
    def from_frame_summary(cls, frame: traceback.FrameSummary):
        return cls(
            filename=frame.filename,
            lineno=frame.lineno,
            name=frame.name,
            line=frame.line,
        )

    def to_frame_summary(self) -> traceback.FrameSummary:
        return traceback.FrameSummary(
            filename=self.filename, lineno=self.lineno, name=self.name, line=self.line
        )


class RemoteErrorDetail(pydantic.BaseModel):
    remote_name: str
    exception_cls_name: str
    exception_module_name: Optional[str]
    exception_message: str
    user_stack_trace: list[StackFrame]

    def to_stack_summary(self) -> traceback.StackSummary:
        return traceback.StackSummary.from_list(
            frame.to_frame_summary() for frame in self.user_stack_trace
        )

    def format(self) -> str:
        stack = "".join(traceback.format_list(self.to_stack_summary()))
        exc_info = (
            f"\n(Exception class defined in `{self.exception_module_name}`.)"
            if self.exception_module_name
            else ""
        )
        error = (
            f"{RemoteErrorDetail.__name__} in `{self.remote_name}`\n"
            f"Traceback (most recent call last):\n"
            f"{stack}{self.exception_cls_name}: {self.exception_message}{exc_info}"
        )
        return error


class GenericRemoteException(Exception):
    ...



import abc
import functools
import logging
from typing import Optional, Type, TypeVar, final

import httpx
from truss_chains import definitions, utils

DEFAULT_TIMEOUT_SEC = 600


class BasetenSession:
    """Helper to invoke predict method on baseten deployments."""

    def __init__(
        self, service_descriptor: definitions.ServiceDescriptor, api_key: str
    ) -> None:
        logging.info(
            f"Creating stub for `{service_descriptor.name}` with predict URL:\n"
            f"\t`{service_descriptor.predict_url}`"
        )
        self._auth_header = {"Authorization": f"Api-Key {api_key}"}
        self._service_descriptor = service_descriptor

    @functools.cached_property
    def _client_sync(self) -> httpx.Client:
        return httpx.Client(headers=self._auth_header, timeout=DEFAULT_TIMEOUT_SEC)

    @functools.cached_property
    def _client_async(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(headers=self._auth_header, timeout=DEFAULT_TIMEOUT_SEC)

    def predict_sync(self, json_payload):
        return utils.handle_response(
            self._client_sync.post(
                self._service_descriptor.predict_url, json=json_payload
            )
        )

    async def predict_async(self, json_payload):
        return utils.handle_response(
            await self._client_async.post(
                self._service_descriptor.predict_url, json=json_payload
            )
        )


class StubBase(abc.ABC):
    _remote: BasetenSession

    @final
    def __init__(
        self, service_descriptor: definitions.ServiceDescriptor, api_key: str
    ) -> None:
        self._remote = BasetenSession(service_descriptor, api_key)

    @classmethod
    def from_url(
        cls,
        predict_url: str,
        context: definitions.DeploymentContext,
        name: Optional[str] = None,
    ):
        name = name or cls.__name__
        return cls(
            definitions.ServiceDescriptor(name=name, predict_url=predict_url),
            api_key=context.get_baseten_api_key(),
        )


StubT = TypeVar("StubT", bound=StubBase)


def factory(
    stub_cls: Type[StubT], context: definitions.DeploymentContext, chainlet_name: str
) -> StubT:
    return stub_cls(
        service_descriptor=context.get_service_descriptor(chainlet_name),
        api_key=context.get_baseten_api_key(),
    )



import collections
import contextlib
import importlib.util
import inspect
import logging
import os
import pathlib
import sys
import types
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Type,
    get_args,
    get_origin,
)

import pydantic
from truss_chains import definitions

_SIMPLE_TYPES = {int, float, complex, bool, str, bytes, None}
_SIMPLE_CONTAINERS = {list, dict}


# Checking of Chainlet class definition ###############################################


def _validate_io_type(param: inspect.Parameter) -> None:
    """
    For Chainlet I/O (both data or parameters), we allow simple types
    (int, str, float...) and `list` or `dict` containers of these.
    Any deeper nested and structured data must be typed as a pydantic model.
    """
    anno = param.annotation
    if isinstance(anno, str):
        raise definitions.APIDefinitionError(
            f"A string-valued type annotation was found: `{param}`. Use only actual "
            "types and avoid 'from __future__ import annotations' (upgrade python)."
        )
    if anno in _SIMPLE_TYPES:
        return
    if isinstance(anno, types.GenericAlias):
        if get_origin(anno) not in _SIMPLE_CONTAINERS:
            raise definitions.APIDefinitionError(
                f"For generic types, only containers {_SIMPLE_CONTAINERS} are "
                f"allowed, but got `{param}`."
            )
        args = get_args(anno)
        for arg in args:
            if arg not in _SIMPLE_TYPES:
                raise definitions.APIDefinitionError(
                    f"For generic types, only arg types {_SIMPLE_TYPES} are "
                    f"allowed, but got `{param}`."
                )
        return
    if issubclass(anno, pydantic.BaseModel):
        # TODO: generating models for the stubs only covers the schema-capabilities
        #   but not any methods that clients might add to their pydantic models.
        #   We should enforce that no user-defined methods (or class vars or properties)
        #   are present, since this might lead to broken behavior.
        # TODO: for enums we rely on the convention that they are string enums and
        #  the member names are the capitalized member values (or better the values are
        #  capitalized themselves too). This constraint should be enforced / checked.
        #  We could also require using a custom StrEnum-like enum base class.
        try:
            anno.schema()
        except Exception as e:
            raise definitions.APIDefinitionError(
                "Pydantic annotations must be able to generate a schema. "
                f"Please fix `{param}`."
            ) from e
        return

    raise definitions.APIDefinitionError(anno)


def _validate_endpoint_params(
    params: list[inspect.Parameter], cls_name: str
) -> list[tuple[str, definitions.TypeDescriptor]]:
    if len(params) == 0:
        raise definitions.APIDefinitionError(
            f"`{cls_name}.{definitions.ENDPOINT_METHOD_NAME}` must be a method, i.e. "
            "with `self` argument."
        )
    if params[0].name != definitions.SELF_ARG_NAME:
        raise definitions.APIDefinitionError(
            f"`{cls_name}.{definitions.ENDPOINT_METHOD_NAME}` must be a method, i.e. "
            "with `self` argument."
        )
    input_name_and_types = []
    for param in params[1:]:  # Skip self argument.
        if param.annotation == inspect.Parameter.empty:
            raise definitions.APIDefinitionError(
                "Inputs of endpoints must have type annotations. "
                f"For `{cls_name}` got:\n{param}"
            )
        _validate_io_type(param)
        type_descriptor = definitions.TypeDescriptor(raw=param.annotation)
        input_name_and_types.append((param.name, type_descriptor))
    return input_name_and_types


def _validate_and_describe_endpoint(
    cls: Type[definitions.ABCChainlet],
) -> definitions.EndpointAPIDescriptor:
    """The "endpoint method" of a Chainlet must have the following signature:

    ```
    [async] def run(
        self, [param_0: anno_0, param_1: anno_1 = default_1, ...]) -> ret_anno:
    ```

    * The name must be `run`.
    * It can be sync or async or def.
    * The number and names of parameters are arbitrary, both positional and named
      parameters are ok.
    * All parameters and the return value must have type annotations. See
      `_validate_io_type` for valid types.
    * Generators are allowed, too (but not yet supported).
    """
    if not hasattr(cls, definitions.ENDPOINT_METHOD_NAME):
        raise definitions.APIDefinitionError(
            f"`{cls.__name__}` must have a {definitions.ENDPOINT_METHOD_NAME}` method."
        )
    endpoint_method = getattr(
        cls, definitions.ENDPOINT_METHOD_NAME
    )  # This is the unbound method.
    if not inspect.isfunction(endpoint_method):
        raise definitions.APIDefinitionError(
            f"`{cls.__name__}.{definitions.ENDPOINT_METHOD_NAME}` must be a method."
        )

    signature = inspect.signature(endpoint_method)
    input_name_and_types = _validate_endpoint_params(
        list(signature.parameters.values()), cls.__name__
    )

    if signature.return_annotation == inspect.Parameter.empty:
        raise definitions.APIDefinitionError(
            f"Return values of endpoints must be type annotated. Got:\n{signature}"
        )
    if get_origin(signature.return_annotation) is tuple:
        output_types = list(
            definitions.TypeDescriptor(raw=arg)
            for arg in get_args(signature.return_annotation)
        )
    else:
        output_types = [definitions.TypeDescriptor(raw=signature.return_annotation)]

    if inspect.isasyncgenfunction(endpoint_method):
        is_async = True
        is_generator = True
    elif inspect.iscoroutinefunction(endpoint_method):
        is_async = True
        is_generator = False
    else:
        is_async = False
        is_generator = inspect.isgeneratorfunction(endpoint_method)

    return definitions.EndpointAPIDescriptor(
        input_names_and_types=input_name_and_types,
        output_types=output_types,
        is_async=is_async,
        is_generator=is_generator,
    )


def _get_generic_class_type(var):
    """Extracts `SomeGeneric` from `SomeGeneric` or `SomeGeneric[T]` uniformly."""
    origin = get_origin(var)
    return origin if origin is not None else var


def _validate_dependency_arg(param) -> Type[definitions.ABCChainlet]:
    # TODO: handle subclasses, unions, optionals, check default value etc.
    if not isinstance(param.default, ChainletProvisionPlaceholder):
        raise definitions.APIDefinitionError(
            f"Any extra arguments of a Chainlet's __init__ must have a default "
            f"value of type `{ChainletProvisionPlaceholder}` (created with the "
            f"`provide` directive). Got `{param.default}` for `{param.name}`."
        )
    chainlet_cls = param.default.chainlet_cls
    if not (
        # TODO: `Protocol` is not a proper class and this might be version dependent.
        # Find a better way to inspect this.
        issubclass(param.annotation, Protocol)  # type: ignore[arg-type]
        or issubclass(chainlet_cls, param.annotation)
    ):
        definitions.APIDefinitionError(
            f"The type annotation for `{param.name}` must either be a `{Protocol}` "
            "or a class/subclass of the Chainlet type used as default value. "
            f"Got `{param.default}`."
        )
    if not issubclass(chainlet_cls, definitions.ABCChainlet):
        raise definitions.APIDefinitionError(
            f"`{chainlet_cls}` must be a subclass of `{definitions.ABCChainlet}`."
        )
    return chainlet_cls


class _ChainletInitParams:
    def __init__(self, params: list[inspect.Parameter]) -> None:
        self._params = params
        self._validate_self_arg()
        self._validate_context_arg()

    def _validate_self_arg(self) -> None:
        if len(self._params) == 0:
            raise definitions.APIDefinitionError(
                "Methods must have first argument `self`."
            )

        if self._params[0].name != definitions.SELF_ARG_NAME:
            raise definitions.APIDefinitionError(
                "Methods must have first argument `self`."
            )

    def _validate_context_arg(self) -> None:
        context_exception = definitions.APIDefinitionError(
            f"`{definitions.ABCChainlet}` must have "
            f"`{definitions.CONTEXT_ARG_NAME}` argument of type "
            f"`{definitions.DeploymentContext}`."
        )
        if len(self._params) < 2:
            raise context_exception
        if self._params[1].name != definitions.CONTEXT_ARG_NAME:
            raise context_exception

        param = self._params[1]
        param_type = _get_generic_class_type(param.annotation)
        if not issubclass(param_type, definitions.DeploymentContext):
            raise context_exception
        if not isinstance(param.default, ContextProvisionPlaceholder):
            raise definitions.APIDefinitionError(
                f"The default value for the `context` argument of a Chainlet's "
                f"__init__ must be of type `{ContextProvisionPlaceholder}` (created "
                f"with the `provide_context` directive). Got `{param.default}`."
            )

    def validated_dependencies(self) -> Mapping[str, Type[definitions.ABCChainlet]]:
        used_classes = set()
        dependencies = {}
        for param in self._params[2:]:  # Skip self and context.
            chainlet_cls = _validate_dependency_arg(param)
            if chainlet_cls in used_classes:
                raise definitions.APIDefinitionError(
                    f"The same Chainlet class cannot be used multiple times for "
                    f"different arguments. Got previously used `{chainlet_cls}` "
                    f"for `{param.name}`."
                )
            dependencies[param.name] = chainlet_cls
            used_classes.add(chainlet_cls)
        return dependencies


def _validate_init_and_get_dependencies(
    cls: Type[definitions.ABCChainlet],
) -> Mapping[str, Type[definitions.ABCChainlet]]:
    """The `__init__`-method of a Chainlet must have the following signature:

    ```
    def __init__(
        self,
        context: truss_chains.Context = truss_chains.provide_context(),
        [dep_0: dep_0_type = truss_chains.provide(dep_0_proc_class),]
        [dep_1: dep_1_type = truss_chains.provide(dep_1_proc_class),]
        ...
    ) -> None:
    ```

    * The context argument is required and must have a default constructed with the
      `provide_context` directive. The type can be templated by a user defined config
      e.g. `truss_chains.Context[UserConfig]`.
    * The names and number of other - "dependency" - arguments are arbitrary.
    * Default values for dependencies must be constructed with the `provide` directive
      to make the dependency injection work. The argument to `provide` must be a
      Chainlet class.
    * The type annotation for dependencies can be a Chainlet class, but it can also be
      a `Protocol` with an equivalent `run` method (e.g. for getting correct type
      checks when providing fake Chainlets for local testing.).
    """
    params = _ChainletInitParams(
        list(inspect.signature(cls.__init__).parameters.values())
    )
    return params.validated_dependencies()


def _validate_variable_access(cls: Type[definitions.ABCChainlet]) -> None:
    # TODO ensure that Chainlets are only accessed via `provided` in `__init__`,`
    # not from manual instantiations on module-level or nested in a Chainlet.
    # See other constraints listed in:
    # https://www.notion.so/ml-infra/WIP-Orchestration-a8cb4dad00dd488191be374b469ffd0a?pvs=4#7df299eb008f467a80f7ee3c0eccf0f0
    ...


def check_and_register_class(cls: Type[definitions.ABCChainlet]) -> None:
    chainlet_descriptor = definitions.ChainletAPIDescriptor(
        chainlet_cls=cls,
        dependencies=_validate_init_and_get_dependencies(cls),
        endpoint=_validate_and_describe_endpoint(cls),
        src_path=os.path.abspath(inspect.getfile(cls)),
        user_config_type=definitions.TypeDescriptor(raw=type(cls.default_user_config)),
    )
    logging.debug(f"Descriptor for {cls}:\n{chainlet_descriptor}\n")
    _validate_variable_access(cls)
    global_chainlet_registry.register_chainlet(chainlet_descriptor)


# Dependency-Injection / Registry ######################################################


class _BaseProvisionPlaceholder:
    """A marker for object to be dependency injected by the framework."""


class ChainletProvisionPlaceholder(_BaseProvisionPlaceholder):
    chainlet_cls: Type[definitions.ABCChainlet]

    def __init__(self, chainlet_cls: Type[definitions.ABCChainlet]) -> None:
        self.chainlet_cls = chainlet_cls

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.chainlet_cls.__name__})"


class ContextProvisionPlaceholder(_BaseProvisionPlaceholder):
    def __str__(self) -> str:
        return f"{self.__class__.__name__}"


class _ChainletRegistry:
    # Because dependencies are required to be present when registering a Chainlet,
    # this dict contains natively a topological sorting of the dependency graph.
    _chainlets: collections.OrderedDict[
        Type[definitions.ABCChainlet], definitions.ChainletAPIDescriptor
    ]
    _name_to_cls: MutableMapping[str, Type]

    def __init__(self) -> None:
        self._chainlets = collections.OrderedDict()
        self._name_to_cls = {}

    def register_chainlet(self, chainlet_descriptor: definitions.ChainletAPIDescriptor):
        for dep in chainlet_descriptor.dependencies.values():
            # To depend on a Chainlet, the class must be defined (module initialized)
            # which entails that is has already been added to the registry.
            assert dep in self._chainlets, dep

        # Because class are globally unique, to prevent re-use / overwriting of names,
        # We must check this in addition.
        if chainlet_descriptor.name in self._name_to_cls:
            conflict = self._name_to_cls[chainlet_descriptor.name]
            existing_source_path = self._chainlets[conflict].src_path
            raise definitions.APIDefinitionError(
                f"A Chainlet with name `{chainlet_descriptor.name}` was already "
                f"defined, Chainlets names must be unique. The pre-existing name "
                f"comes from:\n`{existing_source_path}`\nNew conflicting from\n "
                f"{chainlet_descriptor.src_path}"
            )

        self._chainlets[chainlet_descriptor.chainlet_cls] = chainlet_descriptor
        self._name_to_cls[chainlet_descriptor.name] = chainlet_descriptor.chainlet_cls

    @property
    def chainlet_descriptors(self) -> list[definitions.ChainletAPIDescriptor]:
        return list(self._chainlets.values())

    def get_descriptor(
        self, chainlet_cls: Type[definitions.ABCChainlet]
    ) -> definitions.ChainletAPIDescriptor:
        return self._chainlets[chainlet_cls]

    def get_dependencies(
        self, Chainlet: definitions.ChainletAPIDescriptor
    ) -> Iterable[definitions.ChainletAPIDescriptor]:
        return [
            self._chainlets[desc]
            for desc in self._chainlets[Chainlet.chainlet_cls].dependencies.values()
        ]


global_chainlet_registry = _ChainletRegistry()


# Chainlet class runtime utils #########################################################


def _determine_arguments(func: Callable, **kwargs):
    """Merges provided and default arguments to effective invocation arguments."""
    sig = inspect.signature(func)
    bound_args = sig.bind_partial(**kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments


def ensure_args_are_injected(cls, original_init: Callable, kwargs) -> None:
    """Asserts all placeholder markers are replaced by actual objects."""
    final_args = _determine_arguments(original_init, **kwargs)
    for name, value in final_args.items():
        if isinstance(value, _BaseProvisionPlaceholder):
            raise definitions.UsageError(
                f"When initializing class `{cls.__name__}`, for "
                f"default argument `{name}` a symbolic placeholder value "
                f"was passed (`{value}`). Chainlets must be either a) locally "
                f"instantiated in `{run_local.__name__}` context or b) deployed "
                "remotely. Naive instantiations are prohibited."
            )


# Local Deployment #####################################################################


def _create_local_context(
    chainlet_cls: Type[definitions.ABCChainlet], secrets: Mapping[str, str]
) -> definitions.DeploymentContext:
    return definitions.DeploymentContext(
        user_config=chainlet_cls.default_user_config, secrets=secrets
    )


def _create_modified_init_for_local(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    cls_to_instance: MutableMapping[
        Type[definitions.ABCChainlet], definitions.ABCChainlet
    ],
    secrets: Mapping[str, str],
):
    """Replaces the default argument values with local Chainlet instantiations.

    If this patch is used, Chainlets can be functionally instantiated without
    any init args (because the patched defaults are sufficient).
    """
    original_init = chainlet_descriptor.chainlet_cls.__init__

    def init_for_local(self: definitions.ABCChainlet, **kwargs) -> None:
        logging.debug(f"Patched `__init__` of `{chainlet_descriptor.name}`.")
        kwargs_mod = dict(kwargs)
        if definitions.CONTEXT_ARG_NAME not in kwargs_mod:
            context = _create_local_context(chainlet_descriptor.chainlet_cls, secrets)
            kwargs_mod[definitions.CONTEXT_ARG_NAME] = context
        else:
            logging.debug(
                f"Use explicitly given context for `{self.__class__.__name__}`."
            )
        for arg_name, dep_cls in chainlet_descriptor.dependencies.items():
            if arg_name in kwargs_mod:
                logging.debug(
                    f"Use explicitly given instance for `{arg_name}` of "
                    f"type `{dep_cls.__name__}`."
                )
                continue
            if dep_cls in cls_to_instance:
                logging.debug(
                    f"Use previously created instance for `{arg_name}` of type "
                    f"`{dep_cls.__name__}`."
                )
                instance = cls_to_instance[dep_cls]
            else:
                logging.debug(
                    f"Create new instance for `{arg_name}` "
                    f"of type `{dep_cls.__name__}`."
                )
                assert dep_cls._init_is_patched
                instance = dep_cls()  # type: ignore  # Here init args are patched.
                cls_to_instance[dep_cls] = instance

            kwargs_mod[arg_name] = instance

        original_init(self, **kwargs_mod)

    return init_for_local


@contextlib.contextmanager
def run_local(secrets: Optional[Mapping[str, str]] = None) -> Any:
    """Context to run Chainlets with dependency injection from local instances."""
    type_to_instance: MutableMapping[
        Type[definitions.ABCChainlet], definitions.ABCChainlet
    ] = {}
    original_inits: MutableMapping[Type[definitions.ABCChainlet], Callable] = {}

    for chainlet_descriptor in global_chainlet_registry.chainlet_descriptors:
        original_inits[
            chainlet_descriptor.chainlet_cls
        ] = chainlet_descriptor.chainlet_cls.__init__
        init_for_local = _create_modified_init_for_local(
            chainlet_descriptor, type_to_instance, secrets or {}
        )
        chainlet_descriptor.chainlet_cls.__init__ = init_for_local  # type: ignore[method-assign]
        chainlet_descriptor.chainlet_cls._init_is_patched = True
    try:
        yield
    finally:
        # Restore original classes to unpatched state.
        for chainlet_cls, original_init in original_inits.items():
            chainlet_cls.__init__ = original_init  # type: ignore[method-assign]
            chainlet_cls._init_is_patched = False


########################################################################################


def import_target(
    module_path: pathlib.Path, target_name: str
) -> Type[definitions.ABCChainlet]:
    module_path = pathlib.Path(module_path).resolve()
    module_name = module_path.stem  # Use the file's name as the module name

    error_msg = f"Could not import `{target_name}` from `{module_path}`."
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if not spec:
        raise ImportError(error_msg)
    if not spec.loader:
        raise ImportError(error_msg)

    module = importlib.util.module_from_spec(spec)
    module.__file__ = str(module_path)
    # Since the framework depends on tracking the source files via `inspect` and this
    # depends on the modules bein properly registered in `sys.modules`, we have to
    # manually do this here (because importlib does not do it automatically). This
    # registration has to stay at least until the deployment command has finished.
    # TODO: Might need a context manager to remote this after deploy.
    sys.modules[module_name] = module
    # Add path for making absolute import w.r.t to the source_module's dir.
    sys.path.insert(0, str(module_path.parent))
    try:
        spec.loader.exec_module(module)
    except Exception as e:
        del sys.modules[module_name]
        raise e from e
    finally:
        sys.path.pop(0)

    target_cls = getattr(module, target_name, None)
    if not target_cls:
        raise AttributeError(f"`{target_name}` not found in `{module_path}`.")
    if not issubclass(target_cls, definitions.ABCChainlet):
        raise TypeError(f"{target_cls} is not a {definitions.ABCChainlet}")

    return target_cls



import collections
import inspect
import logging
import os
import pathlib
import shutil
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    cast,
)

import pydantic
import truss
import truss_chains as chains
from truss import truss_config
from truss.contexts.image_builder import serving_image_builder
from truss.remote import remote_cli, remote_factory
from truss.remote.baseten import service as b10_service
from truss_chains import code_gen, definitions, framework, model_skeleton, utils

_REQUIREMENTS_FILENAME = "pip_requirements.txt"
_MODEL_FILENAME = "model.py"
_MODEL_CLS_NAME = model_skeleton.TrussChainletModel.__name__
_TRUSS_DIR = ".truss_gen"


# Truss Gen ############################################################################


def _copy_python_source_files(root_dir: pathlib.Path, dest_dir: pathlib.Path) -> None:
    """Copy all python files under root recursively, but skips generated code."""

    def python_files_only(path, names):
        return [
            name
            for name in names
            if os.path.isfile(os.path.join(path, name))
            and not name.endswith(".py")
            or definitions.GENERATED_CODE_DIR in name
        ]

    shutil.copytree(root_dir, dest_dir, ignore=python_files_only, dirs_exist_ok=True)


def _make_truss_config(
    truss_dir: pathlib.Path,
    chains_config: definitions.RemoteConfig,
    user_config: definitions.UserConfigT,
    chainlet_to_service: Mapping[str, definitions.ServiceDescriptor],
    model_name: str,
) -> truss_config.TrussConfig:
    """Generate a truss config for a Chainlet."""
    config = truss_config.TrussConfig()
    config.model_name = model_name
    config.model_class_filename = _MODEL_FILENAME
    config.model_class_name = _MODEL_CLS_NAME
    # Compute.
    compute = chains_config.get_compute_spec()
    config.resources.cpu = str(compute.cpu)
    config.resources.accelerator = compute.gpu
    config.resources.use_gpu = bool(compute.gpu.count)
    # TODO: expose this setting directly.
    config.runtime.predict_concurrency = compute.cpu
    # Image.
    image = chains_config.get_docker_image_spec()
    config.base_image = truss_config.BaseImage(image=image.base_image)
    pip_requirements: list[str] = []
    if image.pip_requirements_file:
        image.pip_requirements_file.raise_if_not_exists()
        pip_requirements.extend(
            req
            for req in pathlib.Path(image.pip_requirements_file.abs_path)
            .read_text()
            .splitlines()
            if not req.strip().startswith("#")
        )
    pip_requirements.extend(image.pip_requirements)
    # `pip_requirements` will add server requirements which give version conflicts.
    # config.requirements = pip_requirements
    pip_requirements_file_path = truss_dir / _REQUIREMENTS_FILENAME
    pip_requirements_file_path.write_text("\n".join(pip_requirements))
    # TODO: apparently absolute paths don't work with remote build (but work in local).
    config.requirements_file = _REQUIREMENTS_FILENAME  # str(pip_requirements_file_path)
    config.system_packages = image.apt_requirements
    # Assets.
    assets = chains_config.get_asset_spec()
    config.secrets = assets.secrets
    if definitions.BASETEN_API_SECRET_NAME not in config.secrets:
        config.secrets[definitions.BASETEN_API_SECRET_NAME] = definitions.SECRET_DUMMY
    else:
        logging.info(
            f"Chains automatically add {definitions.BASETEN_API_SECRET_NAME} "
            "to secrets - no need to manually add it."
        )
    config.model_cache.models = assets.cached
    # Metadata.
    chains_metadata: definitions.TrussMetadata = definitions.TrussMetadata(
        user_config=user_config, chainlet_to_service=chainlet_to_service
    )
    config.model_metadata[definitions.TRUSS_CONFIG_CHAINS_KEY] = chains_metadata.dict()
    return config


def make_truss(
    chainlet_dir: pathlib.Path,
    chain_root: pathlib.Path,
    chains_config: definitions.RemoteConfig,
    user_config: definitions.UserConfigT,
    model_name: str,
    chainlet_to_service: Mapping[str, definitions.ServiceDescriptor],
    maybe_stub_types_file: Optional[pathlib.Path],
) -> pathlib.Path:
    truss_dir = chainlet_dir / _TRUSS_DIR
    truss_dir.mkdir(exist_ok=True)
    config = _make_truss_config(
        truss_dir, chains_config, user_config, chainlet_to_service, model_name
    )
    config.write_to_yaml_file(
        truss_dir / serving_image_builder.CONFIG_FILE, verbose=True
    )
    # Copy other sources.
    model_dir = truss_dir / truss_config.DEFAULT_MODEL_MODULE_DIR
    model_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy(
        chainlet_dir / f"{definitions.chainlet_MODULE}.py",
        model_dir / _MODEL_FILENAME,
    )
    pkg_dir = truss_dir / truss_config.DEFAULT_BUNDLED_PACKAGES_DIR
    pkg_dir.mkdir(parents=True, exist_ok=True)
    if maybe_stub_types_file is not None:
        shutil.copy(maybe_stub_types_file, pkg_dir)
    # TODO This assume all imports are absolute w.r.t chain root (or site-packages).
    #   Also: apparently packages need an `__init__`, or crash.
    _copy_python_source_files(chain_root, pkg_dir / pkg_dir)

    # TODO Truss package contains this from `{ include = "truss_chains", from = "." }`
    #   pyproject.toml. But for quick dev loop just copy from local.
    shutil.copytree(
        os.path.dirname(chains.__file__),
        pkg_dir / "truss_chains",
        dirs_exist_ok=True,
    )
    return truss_dir


########################################################################################


class DeploymentOptions(pydantic.BaseModel):
    chain_name: str
    only_generate_trusses: bool = False


class DeploymentOptionsBaseten(DeploymentOptions):
    class Config:
        arbitrary_types_allowed = True

    remote_provider: remote_factory.TrussRemote
    publish: bool
    promote: bool

    @classmethod
    def create(
        cls,
        chain_name: str,
        publish: bool,
        promote: bool,
        only_generate_trusses: bool,
        remote: Optional[str] = None,
    ) -> "DeploymentOptionsBaseten":
        if not remote:
            remote = remote_cli.inquire_remote_name(
                remote_factory.RemoteFactory.get_available_config_names()
            )
        return DeploymentOptionsBaseten(
            chain_name=chain_name,
            remote_provider=remote_factory.RemoteFactory.create(remote=remote),
            publish=publish,
            promote=promote,
            only_generate_trusses=only_generate_trusses,
        )


class DeploymentOptionsLocalDocker(DeploymentOptions):
    # Local docker-to-docker requests don't need auth, but we need to set a
    # value different from `SECRET_DUMMY` to not trigger the check that the secret
    # is unset. Additionally, if local docker containers make calls to models deployed
    # on baseten, a real API key must be provided (i.e. the default must be overridden).
    baseten_chain_api_key: str = "docker_dummy_key"


def _deploy_to_baseten(
    truss_dir: pathlib.Path, options: DeploymentOptionsBaseten
) -> b10_service.BasetenService:
    truss_handle = truss.load(str(truss_dir))
    model_name = truss_handle.spec.config.model_name
    assert model_name is not None
    logging.info(
        f"Deploying truss model to Baseten`{model_name}` "
        f"(publish={options.publish}, promote={options.promote})."
    )
    # Models must be trusted to use the API KEY secret.
    service = options.remote_provider.push(
        truss_handle,
        model_name=model_name,
        trusted=True,
        publish=options.publish,
        promote=options.promote,
    )
    if not isinstance(service, b10_service.BasetenService):
        raise TypeError("Remote provider did not return baseten service.")
    return cast(b10_service.BasetenService, service)


class DockerService(b10_service.TrussService):
    def __init__(self, remote_url: str, is_draft: bool, **kwargs):
        super().__init__(remote_url, is_draft, **kwargs)

    def authenticate(self) -> Dict[str, str]:
        return {}

    def is_live(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    def is_ready(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    @property
    def logs_url(self) -> str:
        raise NotImplementedError()

    @property
    def predict_url(self) -> str:
        return f"{self._service_url}/v1/models/model:predict"

    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        raise NotImplementedError()


def _deploy_service(
    chainlet_dir: pathlib.Path,
    chain_root: pathlib.Path,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    chainlet_to_service: Mapping[str, definitions.ServiceDescriptor],
    maybe_stub_types_file: Optional[pathlib.Path],
    options: DeploymentOptions,
) -> tuple[Optional[b10_service.TrussService], definitions.ServiceDescriptor]:
    # Filter only needed services.
    chainlet_to_service = {
        dep_proc.__name__: chainlet_to_service[dep_proc.__name__]
        for dep_proc in chainlet_descriptor.dependencies.values()
    }
    # Convert to truss and deploy.
    # TODO: support file-based config (and/or merge file and python-src config values).
    remote_config = chainlet_descriptor.chainlet_cls.remote_config
    chainlet_name = remote_config.name or chainlet_descriptor.name
    model_name = f"{options.chain_name}.{chainlet_name}"
    truss_dir = make_truss(
        chainlet_dir,
        chain_root,
        remote_config,
        chainlet_descriptor.chainlet_cls.default_user_config,
        model_name,
        chainlet_to_service,
        maybe_stub_types_file,
    )
    service: Optional[b10_service.TrussService]
    if options.only_generate_trusses:
        service_descr = definitions.ServiceDescriptor(
            name=model_name, predict_url="https://dummy"
        )
        service = None
    elif isinstance(options, DeploymentOptionsLocalDocker):
        port = utils.get_free_port()
        truss_handle = truss.load(str(truss_dir))
        truss_handle.add_secret(
            definitions.BASETEN_API_SECRET_NAME, options.baseten_chain_api_key
        )
        _ = truss_handle.docker_run(
            local_port=port, detach=True, wait_for_server_ready=True, network="host"
        )

        service_descr = definitions.ServiceDescriptor(
            name=model_name,
            # Localhost seems to only work *sometimes* with docker.
            # predict_url=f"http://localhost:{port}/v1/models/model:predict",
            predict_url=f"http://host.docker.internal:{port}/v1/models/model:predict",
        )
        service = DockerService(f"http://localhost:{port}", is_draft=True)

    elif isinstance(options, DeploymentOptionsBaseten):
        with utils.log_level(logging.INFO):
            service = _deploy_to_baseten(truss_dir, options)
        service_descr = definitions.ServiceDescriptor(
            name=model_name, predict_url=service.predict_url
        )
    else:
        raise NotImplementedError(options)

    logging.info(f"Deployed service: {service_descr}.")
    return service, service_descr


def _get_ordered_dependencies(
    Chainlets: Iterable[Type[definitions.ABCChainlet]],
) -> Iterable[definitions.ChainletAPIDescriptor]:
    """Gather all Chainlets needed and returns a topologically ordered list."""
    needed_chainlets: set[definitions.ChainletAPIDescriptor] = set()

    def add_needed_chainlets(proc: definitions.ChainletAPIDescriptor):
        needed_chainlets.add(proc)
        for chainlet_descriptor in framework.global_chainlet_registry.get_dependencies(
            proc
        ):
            needed_chainlets.add(chainlet_descriptor)
            add_needed_chainlets(chainlet_descriptor)

    for proc_cls in Chainlets:
        add_needed_chainlets(
            framework.global_chainlet_registry.get_descriptor(proc_cls)
        )

    # Iterating over the registry ensures topological ordering.
    return [
        chainlet_descriptor
        for chainlet_descriptor in framework.global_chainlet_registry.chainlet_descriptors
        if chainlet_descriptor in needed_chainlets
    ]


class ChainService:
    _entrypoint: str
    _services: MutableMapping[str, b10_service.TrussService]

    def __init__(self, entrypoint: str, name: str) -> None:
        self.name = name
        self._entrypoint = entrypoint
        self._services = collections.OrderedDict()  # Preserve order.

    def add_service(self, name: str, service: b10_service.TrussService) -> None:
        self._services[name] = service

    @property
    def get_entrypoint(self) -> b10_service.TrussService:
        service = self._services.get(self._entrypoint)
        if not service:
            raise definitions.MissingDependencyError(
                f"Service for entrypoint `{self._entrypoint}` was not added."
            )
        return service

    @property
    def run_url(self) -> str:
        return self.get_entrypoint.predict_url

    def run(self, json: Dict) -> Any:
        return self.get_entrypoint.predict(json)

    def get_info(self) -> list[tuple[str, str, str]]:
        """Return list with elements (name, status, logs_url) for each chainlet."""
        return list(
            (name, next(service.poll_deployment_status(sleep_secs=0)), service.logs_url)
            for name, service in self._services.items()
        )


def deploy_remotely(
    entrypoint: Type[definitions.ABCChainlet],
    options: DeploymentOptions,
    non_entrypoint_root_dir: Optional[str] = None,
) -> ChainService:
    """
    * Gathers dependencies of `entrypoint`.
    * Generates stubs.
    * Generates truss model code, including stub initialization.
    * Deploys truss models to baseten.
    """
    # TODO: revisit how chain root is inferred/specified, current might be brittle.
    if non_entrypoint_root_dir:
        chain_root = pathlib.Path(non_entrypoint_root_dir).absolute()
    else:
        chain_root = pathlib.Path(inspect.getfile(entrypoint)).absolute().parent
    logging.info(f"Using project root for chain: `{chain_root}`.")

    chainlet_to_service: dict[str, definitions.ServiceDescriptor] = {}
    chain_service = ChainService(
        framework.global_chainlet_registry.get_descriptor(entrypoint).name,
        name=options.chain_name,
    )
    for chainlet_descriptor in _get_ordered_dependencies([entrypoint]):
        logging.info(f"Deploying `{chainlet_descriptor.name}`.")
        deps = framework.global_chainlet_registry.get_dependencies(chainlet_descriptor)
        chainlet_dir = code_gen.make_chainlet_dir(
            options.chain_name, chainlet_descriptor
        )
        chainlet_filepath = pathlib.Path(
            shutil.copy(
                chainlet_descriptor.src_path,
                chainlet_dir / f"{definitions.chainlet_MODULE}.py",
            )
        )
        maybe_stub_types_file = code_gen.gen_pydantic_models(
            chainlet_dir / f"{definitions.STUB_TYPE_MODULE}.py", deps
        )
        code_gen.generate_chainlet_source(
            chainlet_filepath, chainlet_descriptor, deps, maybe_stub_types_file
        )
        service, service_descr = _deploy_service(
            chainlet_dir,
            chain_root,
            chainlet_descriptor,
            chainlet_to_service,
            maybe_stub_types_file,
            options,
        )
        if service:
            chain_service.add_service(chainlet_descriptor.name, service)
        chainlet_to_service[chainlet_descriptor.name] = service_descr

    return chain_service



from typing import Any, ContextManager, Mapping, Optional, Type, final

from truss_chains import definitions, deploy, framework


def provide_context() -> Any:
    """Sets a 'symbolic marker' for injecting a Context object at runtime."""
    return framework.ContextProvisionPlaceholder()


def provide(chainlet_cls: Type[definitions.ABCChainlet]) -> Any:
    """Sets a 'symbolic marker' for injecting a stub or local Chainlet at runtime."""
    # TODO: extend with RPC customization, e.g. timeouts, retries etc.
    return framework.ChainletProvisionPlaceholder(chainlet_cls)


class ChainletBase(definitions.ABCChainlet[definitions.UserConfigT]):
    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        framework.check_and_register_class(cls)

        original_init = cls.__init__

        def init_with_arg_check(self, *args, **kwargs):
            if args:
                raise definitions.UsageError("Only kwargs are allowed.")
            framework.ensure_args_are_injected(cls, original_init, kwargs)
            original_init(self, *args, **kwargs)

        cls.__init__ = init_with_arg_check  # type: ignore[method-assign]

    def __init__(
        self,
        context: definitions.DeploymentContext[
            definitions.UserConfigT
        ] = provide_context(),
    ) -> None:
        self._context = context

    @final
    @property
    def user_config(self) -> definitions.UserConfigT:
        return self._context.user_config


def deploy_remotely(
    entrypoint: Type[definitions.ABCChainlet],
    chain_name: str,
    publish: bool = True,
    promote: bool = True,
    only_generate_trusses: bool = False,
) -> deploy.ChainService:
    options = deploy.DeploymentOptionsBaseten.create(
        chain_name=chain_name,
        publish=publish,
        promote=promote,
        only_generate_trusses=only_generate_trusses,
    )
    return deploy.deploy_remotely(entrypoint, options)


def run_local(secrets: Optional[Mapping[str, str]] = None) -> ContextManager[None]:
    """Context manager for using in-process instantiations of Chainlet dependencies."""
    return framework.run_local(secrets)
